;; Generated package description from disk-usage.el  -*- no-byte-compile: t -*-
(define-package "disk-usage" "1.3.3.0.20230920.164444" "Sort and browse disk usage listings" '((emacs "26.1")) :commit "b0d803f2cec3afc2937840f9ba66e3f903d6c415" :authors '(("Pierre Neidhardt" . "mail@ambrevar.xyz")) :maintainer '("Pierre Neidhardt" . "mail@ambrevar.xyz") :keywords '("files" "convenience" "tools") :url "https://gitlab.com/Ambrevar/emacs-disk-usage")
